from __future__ import annotations

import contextlib
import dataclasses as dc
import logging
import shutil
import tempfile
from pathlib import Path
from typing import Generator

from .run1 import (
    EMode,
    OMode,
    Paths,
    RunnerError,  # noqa: F401
    mkpaths,
    runc,
)

logger = logging.getLogger(__name__)


@dc.dataclass
class Runner:
    verbose: bool
    dryrun: bool | None = None
    exe: Paths | None = None
    cwd: Path | None = None
    overrides: dict[str, str] | None = None
    log: logging.Logger | None = None

    @staticmethod
    @contextlib.contextmanager
    def tmpdir(source: Path | None = None) -> Generator[Path, None, None]:
        wdir = source if source else Path(tempfile.mkdtemp())
        wdir.mkdir(parents=True, exist_ok=True)
        try:
            yield wdir
        finally:
            if not source:
                shutil.rmtree(wdir, ignore_errors=True)

    def __call__(
        self,
        args: Paths,
        capture: bool = False,
        verbose: bool | None = None,
        dryrun: bool | None = None,
        cwd: Path | str | bool | None = None,
        overrides: dict[str, str] | None = None,
        log: logging.Logger | None = None,
    ) -> str | bytes | None:
        # capture/verbose are interacting to control how the stdout is handled
        check = (capture, self.verbose if verbose is None else verbose)
        stdout: OMode = "null"
        if check == (True, False):
            stdout = "capture"
        elif check == (True, True):
            stdout = "capture+display"
        elif check == (False, True):
            stdout = "display"
        elif check == (False, False):
            stdout = "null"
        else:
            raise RuntimeError(f"un-handled value {check=}")
        stderr: EMode = "display" if self.verbose else "null"

        dryrun = self.dryrun if dryrun is None else dryrun
        if "capture" in stdout and dryrun:
            raise RuntimeError("cannot dryrun and caputure")

        cwd = cwd or self.cwd
        overrides = overrides or self.overrides
        log = log or self.log or logger

        fullargs = mkpaths(args)
        if self.exe:
            fullargs = [*mkpaths(self.exe), *fullargs]

        log.debug("%srun: %s", "(dry-run) " if dryrun else "", " ".join(fullargs))
        if dryrun:
            return None
        return runc(fullargs, stdout=stdout, stderr=stderr, overrides=overrides, cwd=cwd)


@dc.dataclass
class DryRunner(Runner):
    def __call__(
        self,
        args: Paths,
        capture: bool = False,
        verbose: bool | None = None,
        dryrun: bool | None = None,
        cwd: Path | str | bool | None = None,
        overrides: dict[str, str] | None = None,
        log: logging.Logger | None = None,
    ) -> str | bytes | None:
        if capture:
            raise RuntimeError("cannot dryrun and caputure")
        cwd = cwd or self.cwd
        overrides = overrides or self.overrides
        log = log or self.log or logger

        fullargs = mkpaths(args)
        if self.exe:
            fullargs = [*mkpaths(self.exe), *fullargs]

        (log.info if self.verbose else log.debug)("%srun: %s", "(dry-run) " if dryrun else "", " ".join(fullargs))
        return None


if __name__ == "__main__":
    runner = Runner(True)
    y = runner(["ls", "-l"], capture=False, verbose=True)
    print(y)

    # --git-dir=$dest/.git --work-tree $dest
    runner = Runner(True, exe=["git", "--git-dir", "{workdir}/.git"])
    runner("status")



from __future__ import annotations

import argparse
import inspect
import sys
from typing import Callable

# The return type of add_arguments

if sys.version_info >= (3, 10):
    ArgsCallback = Callable[[argparse.Namespace], None | argparse.Namespace]
else:
    ArgsCallback = Callable


class CliBaseError(Exception):
    pass


class AbortCliError(CliBaseError):
    pass


class AbortWrongArgumentError(CliBaseError):
    pass


def check_default_constructor(klass: type):
    signature = inspect.signature(klass.__init__)  # type: ignore[misc]
    for name, value in signature.parameters.items():
        if name in {"self", "args", "kwargs"}:
            continue
        if value.default is inspect.Signature.empty:
            raise RuntimeError(f"the {klass}() cannot be called without arguments")



import copy

from acbox import dictionaries


def roundtrip(data, updates):
    original = copy.deepcopy(data)
    undo = dictionaries.dictupdate(data, updates)
    dictionaries.dictrollback(data, undo)
    assert data == original


def test_first():
    original = {}
    updates = [
        ("a.b.c.d.e", 1),
    ]

    data = copy.deepcopy(original)
    undo = dictionaries.dictupdate(data, updates)
    assert undo == [
        ("a", dictionaries.DELETE),
        ("a.b", dictionaries.DELETE),
        ("a.b.c", dictionaries.DELETE),
        ("a.b.c.d", dictionaries.DELETE),
        ("a.b.c.d.e", dictionaries.DELETE),
    ]
    assert data == {
        "a": {
            "b": {
                "c": {
                    "d": {
                        "e": 1,
                    }
                }
            }
        }
    }
    dictionaries.dictrollback(data, undo)
    assert data == original


def test_second():
    data = {
        "a": 12,
        "c": {
            "x": 99,
        },
    }

    updates = [
        ("a", 22),
        ("d.e", 33),
        ("c.y.z", 44),
        ("c.x", 55),
    ]

    original = copy.deepcopy(data)
    undo = dictionaries.dictupdate(data, updates)
    assert undo == [
        ("a", 12),
        ("d", dictionaries.DELETE),
        ("d.e", dictionaries.DELETE),
        ("c.y", dictionaries.DELETE),
        ("c.y.z", dictionaries.DELETE),
        ("c.x", 99),
    ]
    assert data == {
        "a": 22,
        "c": {
            "y": {
                "z": 44,
            },
            "x": 55,
        },
        "d": {
            "e": 33,
        },
    }

    dictionaries.dictrollback(data, undo)
    assert data == original


def test_third():
    data = {
        "a": 12,
        "c": {
            "x": 99,
        },
    }

    updates = [
        ("c.y.z", 44),
        ("c.x", dictionaries.DELETE),
    ]

    original = copy.deepcopy(data)
    undo = dictionaries.dictupdate(data, updates)
    assert data == {
        "a": 12,
        "c": {
            "y": {
                "z": 44,
            },
        },
    }
    dictionaries.dictrollback(data, undo)
    assert data == original



import dataclasses as dc
import os
import sys
import time
from typing import Any, BinaryIO

import pytest

from acbox import run1


@dc.dataclass
class BufferFilter(run1.BaseFilter):
    buffer: list[str] = dc.field(default_factory=list)
    exception: Any = ""

    def __call__(self, stream: BinaryIO) -> None:
        try:
            for buffer in iter(stream.readline, b""):
                line = buffer.decode("utf-8")
                if line.endswith(os.linesep):
                    line = line[: -len(os.linesep)]
                self.buffer.append(line)
        except Exception as exc:
            self.exception = exc


@pytest.fixture(scope="function")
def filters():
    @dc.dataclass
    class Filters:
        stdout: BufferFilter  # type: ignore
        stderr: BufferFilter  # type: ignore

        def reset(self):
            self.stdout.buffer = []
            self.stderr.buffer = []

    yield Filters(stdout=BufferFilter(), stderr=BufferFilter())


def test_new_filter(resolver, filters):
    """simply capture the stdout / stderr"""
    cmd = [sys.executable, resolver.lookup("simple-script.py")]

    run1.runc(cmd, **filters.__dict__)
    assert ["1, got 'first message' (<stdout>)"] == filters.stdout.buffer
    assert ["2, got 'second message' (<stderr>)"] == filters.stderr.buffer

    filters.reset()
    run1.runc(cmd, overrides={"VALUE": "123"}, **filters.__dict__)
    assert ["1, got 'first message' (<stdout>)", "3, got 'received [123]' (<stdout>)"] == filters.stdout.buffer
    assert ["2, got 'second message' (<stderr>)", "4, got 'received [123]' (<stderr>)"] == filters.stderr.buffer


def test_new_filter_with_error(resolver, filters):
    cmd = [sys.executable, resolver.lookup("simple-script.py")]
    with pytest.raises(run1.RunnerError) as exc:
        run1.runc(cmd, overrides={"ABORT": "33"}, **filters.__dict__)

    assert ["2, got 'second message' (<stderr>)", "4, got 'aborting with 33' (<stderr>)"] == filters.stderr.buffer
    assert ["1, got 'first message' (<stdout>)", "3, got 'aborting with 33' (<stdout>)"] == filters.stdout.buffer

    assert None is exc.value.cwd
    assert {"ABORT": "33"} == exc.value.overrides
    assert " ".join(str(c) for c in cmd) == exc.value.cmdline
    assert 33 == exc.value.returncode


def test_wait_and_timeout(resolver, filters):
    cmd = [sys.executable, resolver.lookup("simple-script.py")]

    t0 = time.monotonic()
    run1.runc(cmd, overrides={"WAIT": "0.5"}, **filters.__dict__)
    delta = time.monotonic() - t0
    assert delta > 0.5
    assert [
        "1, got 'first message' (<stdout>)",
        "3, got 'waiting 0.5s ..' (<stdout>)",
        "5, got 'done waiting 0.5s' (<stdout>)",
    ] == filters.stdout.buffer
    assert [
        "2, got 'second message' (<stderr>)",
        "4, got 'waiting 0.5s ..' (<stderr>)",
        "6, got 'done waiting 0.5s' (<stderr>)",
    ] == filters.stderr.buffer

    filters.reset()
    with pytest.raises(run1.TimeoutRunnerError) as exc:
        run1.runc(cmd, overrides={"WAIT": "100"}, timeout=0.5, **filters.__dict__)
    assert exc.value.returncode
    assert None is exc.value.cwd
    assert {"WAIT": "100"} == exc.value.overrides
    assert " ".join(str(c) for c in cmd) == exc.value.cmdline



from __future__ import annotations

import dataclasses as dc
from pathlib import Path

from ..runner import Paths, Runner


@dc.dataclass
class GitBranch:
    name: str
    short: str


@dc.dataclass
class Git:
    worktree: Path
    runner: Runner
    gitdir: Path | None = None
    exe: str | Path = "git"

    def __post_init__(self):
        if not self.gitdir:
            self.gitdir = self.worktree / ".git"

    def __repr__(self):
        return f"<{self.__class__.__name__} worktree={self.worktree}>"

    @classmethod
    def new(cls, worktree: Path, verbose: bool = False) -> Git:
        worktree = worktree.expanduser().absolute()
        runner = Runner(verbose=verbose, exe=["git", "--git-dir", f"{worktree}/.git"])
        return cls(worktree, runner=runner)

    @classmethod
    def clone(cls, url: str, worktree: Path | None = None, verbose: bool = False, *args) -> Git:
        git = cls.new(worktree or Path.cwd(), verbose=verbose)
        git.worktree.parent.mkdir(parents=True, exist_ok=True)
        git.runner(["clone", *(args or []), url, git.worktree])
        return git

    def __call__(self, args: Paths) -> str:
        out = self.runner(args, capture=True) or ""
        return (out.decode("utf-8") if isinstance(out, bytes) else out).strip()

    # from here all "command" wrappers

    def remotes(self) -> dict[str, dict[str, str | set[str]]]:
        # origin	git@github.com:cav71/lektor.git (fetch)
        result: dict[tuple[str, str], set[str]] = {}
        out = self(
            [
                "remote",
                "-v",
            ]
        )
        for line in out.strip().split("\n"):
            if len(elems := line.strip().split()) != 3:
                continue
            name, url, action = elems
            if (key := (name, url)) not in result:
                result[key] = set()
            result[key].add(action)

        return {n: {"origin": u, "actions": a} for (n, u), a in result.items()}

    def branch(self, name: str = "HEAD") -> str:
        return self(["rev-parse", "--abbrev-ref", name])

    def commits_on_branch(self, main: str = "origin/main") -> int:
        return int(self(["rev-list", "--count", main]))



import functools
import logging
import sys
from argparse import Namespace
from typing import Any, Callable, Sequence

import click
import cloup
from rich.console import Console
from rich.logging import RichHandler

MainFn = Callable[[Namespace], int | None]
AddArgFn = Callable[[MainFn], MainFn]
ProcessOptionsFn = Callable[[Namespace], None | Namespace]


def add_loglevel(fn: MainFn) -> MainFn:
    fn = cloup.option_group("loglevel", click.option("-q", "--quiet", count=True), click.option("-v", "--verbose", count=True))(fn)
    return fn


def _resolve_log_level(options: Namespace) -> int:
    verbose = (options.__dict__.pop("verbose") if "verbose" in options.__dict__ else 0) - (
        options.__dict__.pop("quiet") if "quiet" in options.__dict__ else 0
    )
    verbose = max(min(verbose, 1), -1)
    level = {-1: logging.WARNING, 0: logging.INFO, 1: logging.DEBUG}[verbose]
    options.verbose = level > logging.INFO
    return level


def process_loglevel(options: Namespace) -> Namespace:
    level = _resolve_log_level(options)
    logging.basicConfig(level=level)
    return options


def process_loglevel_fancy(options: Namespace, verbose_flag: bool = False) -> Namespace:
    level = _resolve_log_level(options)
    stderr = Console(stderr=True)
    logging.basicConfig(
        level=level,
        format="%(asctime)s %(message)s",
        handlers=[RichHandler(console=stderr, show_time=False, markup=True, rich_tracebacks=True)],
    )
    return options


CLICKWRAPPERS: dict[str, tuple[AddArgFn | None, ProcessOptionsFn | None]] = {
    "default": (add_loglevel, process_loglevel),
    "fancy": (add_loglevel, process_loglevel_fancy),
}


def validate_add_argument_function(wrapper: AddArgFn, fn: MainFn) -> MainFn:
    fn1 = wrapper(fn)
    if not callable(fn1):
        raise RuntimeError(f"wrapper function {wrapper} doesn't return a callable")
    return fn1


def clickwrapper(
    add_arguments: AddArgFn | Sequence[AddArgFn] | None = None,
    process_options: ProcessOptionsFn | Sequence[ProcessOptionsFn] | None = None,
) -> Callable[[MainFn], None]:
    def _clickwrapper(fn: MainFn):
        if add_arguments:
            wrappers = add_arguments if isinstance(add_arguments, Sequence) else [add_arguments]
            for wrapper in wrappers:
                fn = validate_add_argument_function(wrapper, fn)

        @functools.wraps(fn)
        def __clickwrapper(*args, **kwargs):
            options = Namespace(**kwargs)

            def error(msg):
                raise click.UsageError(msg)

            def exit(retcode: int = 0):
                sys.exit(retcode)

            if hasattr(options, "error") or hasattr(options, "exit"):
                raise RuntimeError("you have an error|exit option")
            options.error = error
            options.exit = sys.exit

            if process_options:
                processors = process_options if isinstance(process_options, Sequence) else [process_options]
                for process in processors:
                    options = process(options) or options

            if ret := fn(options):
                raise click.exceptions.Exit(ret)
            return ret

        return __clickwrapper

    return _clickwrapper


def clickwrap(
    kind: str | None = "default", add_arguments: AddArgFn | None = None, process_options: ProcessOptionsFn | None = None
) -> Callable[[MainFn], Any]:
    args0: list[AddArgFn] = []
    args1: list[ProcessOptionsFn] = []
    if kind:
        if kind not in CLICKWRAPPERS:
            raise RuntimeError(f"cannot find CLICKWRAPPERS[{kind}]")
        if arg := CLICKWRAPPERS[kind][0]:
            args0.append(arg)
        if arg1 := CLICKWRAPPERS[kind][1]:
            args1.append(arg1)
    if add_arguments:
        args0.append(add_arguments)
    if process_options:
        args1.append(process_options)
    return clickwrapper(args0, args1)


def command(
    kind: str | None = "default", add_arguments: AddArgFn | None = None, process_options: ProcessOptionsFn | None = None
) -> Callable[[MainFn], Any]:
    def decorator(func: MainFn) -> Callable[[MainFn], None]:
        adeco = clickwrap(kind, add_arguments, process_options)(func)
        if kind in {"fancy"}:
            bdeco = cloup.command(
                formatter_settings=cloup.HelpFormatter.settings(theme=cloup.HelpTheme.dark()), context_settings={"show_default": True}
            )(adeco)
        else:
            bdeco = cloup.command()(adeco)
        return bdeco

    return decorator


def group(kind: str | None = "default", add_arguments: AddArgFn | None = None, process_options: ProcessOptionsFn | None = None):
    def _fn(fn):
        g = cloup.group(formatter_settings=cloup.HelpFormatter.settings(theme=cloup.HelpTheme.dark()), show_subcommand_aliases=True)
        group = g(fn)
        group._command = group.command
        group.command = functools.partial(
            group._command,
            formatter_settings=cloup.HelpFormatter.settings(theme=cloup.HelpTheme.dark()),
            context_settings={"show_default": True},
        )
        return group

    return _fn


if __name__ == "__main__":
    log = logging.getLogger(__name__)

    # @cloup.command(formatter_settings=cloup.HelpFormatter.settings(theme=cloup.HelpTheme.dark()))
    @command()
    @clickwrap("default")
    def main(args: Namespace) -> int:
        log.debug("a debug message")
        log.info("an info message, got verbose=%i", args.verbose)
        log.warning("a warning message")
        return 1

    main()



from acbox.services.git import Git


def test_new(tmp_path):
    git = Git.new(tmp_path / "wow")
    assert git.gitdir == (git.worktree / ".git")
    assert not git.worktree.exists()
    assert not git.worktree.exists()


def test_clone(tmp_path):
    git = Git.clone("https://github.com/cav71/acbox.git", tmp_path / "wow1")

    assert git.worktree.exists() and git.worktree.is_dir()
    assert git.gitdir.exists() and git.gitdir.is_dir()
    assert git.gitdir.relative_to(tmp_path)
    assert git.gitdir == (git.worktree / ".git")
    assert (git.worktree / "support/builder.py").exists()
    assert git.branch() == "main"



import inspect
import os
import platform
import shutil
import sys
from pathlib import Path

from acbox.ureporting import Record, S, check, print_report


@check
def check_sys(group) -> list[Record]:
    return [
        Record(S.NOSTATUS, group, "executable", sys.executable),
        Record(S.NOSTATUS, group, "version", str(sys.version_info)),
    ]


@check
def check_plaform(group: str) -> list[Record]:
    return [
        Record(S.NOSTATUS, group, "arch", str(platform.architecture(sys.executable))),
        Record(S.NOSTATUS, group, "system", str(platform.uname().system)),
    ]


@check
def check_environ(group: str) -> list[Record]:
    def chunk(txt, n):
        return [txt[i : i + n] for i in range(0, len(txt), n)]

    special = {
        "PATH": lambda key, value: value.split(os.pathsep),
        "MANPATH": lambda key, value: value.split(os.pathsep),
        "DIRENV_DIFF": lambda key, value: chunk(value, 70),
        "DIRENV_WATCHES": lambda key, value: chunk(value, 70),
        "LS_COLORS": lambda key, value: chunk(value, 70),
        lambda key: key.startswith("GITHUB_"): None,
    }
    for key, fn in special.items():
        if callable(fn) and set(inspect.signature(fn).parameters) != {"key", "value"}:
            raise RuntimeError(f"invalid signature for function at {key=}")
    result = []
    for key, value in sorted(os.environ.items(), key=lambda k: k[0].upper()):
        if key in {"_"}:
            continue
        for keyfn, valuefn in special.items():
            if (callable(keyfn) and keyfn(key)) or (keyfn == key):
                value = valuefn(key, value) if callable(valuefn) else valuefn  # type: ignore
            else:
                continue
        if value is None:
            continue
        result.append(Record(S.NOSTATUS, group, key, value))
    return result


@check
def check_executables(group: str, paths: list[str | Path] | None = None) -> list[Record]:
    exes = paths or [
        "git",
        "python",
        "python3",
        "pip",
        "pip3",
    ]
    result = []
    for exe in exes:
        if found := shutil.which(exe):
            bins = [f"found in {found}"]
            if found != str(Path(found).resolve()):
                bins = [*bins, f"({Path(found).resolve()})"]
            result.append(Record(S.NOSTATUS, group, str(exe), bins))
        else:
            result.append(Record(S.NOSTATUS, group, str(exe), "not found"))
    return result


@check
def check_envfile(group: str) -> list[Record]:
    path = Path("/etc/env.sh")
    result = []
    if path.exists():
        lines = [
            [l1.strip() for l1 in line.split("=")]
            for line in path.read_text().split("\n")
            if line.strip() and "=" in line and len(line.split("=")) == 2
        ]
        for key, value in lines:
            result.append(Record(S.NOSTATUS, group, key, value))
    else:
        result.append(Record(S.NOSTATUS, group, "not-found"))
    return result


def main() -> int:
    report = []
    report.extend(check_sys("sys"))
    report.extend(check_plaform("platform"))
    report.extend(check_environ("environ.env"))
    report.extend(check_executables("environ.exe"))
    report.extend(check_envfile("envfile"))
    return print_report(report)


if __name__ == "__main__":
    sys.exit(main())



import argparse
import logging
from typing import Any

from ..parsers.base import ArgumentParserBase


def add_arguments_logging(parser: ArgumentParserBase, baselevel: int = logging.INFO) -> None:
    group = parser.add_argument_group("Logging", "Logging related options")
    group.add_argument("-v", "--verbose", dest="managed-loglevel", action="append_const", const=1, help="report verbose logging")
    group.add_argument("-q", "--quiet", dest="managed-loglevel", action="append_const", const=-1, help="report quiet logging")

    levelmap = [
        logging.DEBUG,
        logging.INFO,
        logging.WARNING,
        logging.ERROR,
        logging.CRITICAL,
        logging.FATAL,
    ]
    if baselevel not in levelmap:
        raise IndexError(f"cannot find level {baselevel} in: {levelmap}")

    def setup_logging(config: dict[str, Any]) -> None:
        logging.basicConfig(**config)

    def callback(args: argparse.Namespace):
        config = {}
        count = levelmap.index(baselevel) - sum(getattr(args, "managed-loglevel") or [0])
        config["level"] = levelmap[min(max(count, 0), len(levelmap) - 1)]
        setup_logging(config)
        delattr(args, "managed-loglevel")

    parser.callbacks.append(callback)



import sys
from pathlib import Path

from acbox import fileops


def test_which():
    exe = "cmd" if sys.platform == "win32" else "sh"

    path = fileops.which(exe)
    assert path
    assert isinstance(path, Path)

    paths = fileops.which_n(exe)
    assert isinstance(paths, list)
    assert paths[0] == fileops.which(exe)

    assert fileops.which_n("xwdwxEW") is None



from __future__ import annotations

import dataclasses as dc
import datetime
import logging
import os
import subprocess
import sys
import threading
import time
from pathlib import Path
from typing import BinaryIO, Literal, Sequence

COLORS = {
    "blue": "\033[94m",
    "green": "\033[92m",
    "red": "\033[91m",
    "clear": "\033[0m",
}

POLL_S = 0.03

logger = logging.getLogger(__name__)


@dc.dataclass
class RunnerError(Exception):
    message: str
    cwd: Path | None
    overrides: dict[str, str] | None
    cmdline: str
    returncode: int


@dc.dataclass
class TimeoutRunnerError(RunnerError):
    pass


@dc.dataclass
class BaseFilter:
    def __call__(self, stream: BinaryIO) -> None:
        for line in iter(stream.readline, b""):
            pass


OMode = Literal["capture", "null", "display", "capture+display"] | BaseFilter
EMode = Literal["null", "display"] | BaseFilter
Paths = str | Path | Sequence[str | Path]


@dc.dataclass
class CaptureFilter(BaseFilter):
    encode: str | None = "utf-8"
    result: str | bytes | None = None

    def __call__(self, stream: BinaryIO) -> None:
        result = []
        for line in iter(stream.readline, b""):
            result.append(line[:-1])
        stream.close()
        if self.encode:
            self.result = b"\n".join(result).decode(self.encode)
        else:
            self.result = b"\n".join(result)


@dc.dataclass
class DisplayFilter(BaseFilter):
    color: str | None
    pre: str = "   | "
    clear: str = COLORS["clear"]
    capture: bool = False
    encode: str | None = "utf-8"
    result: str | bytes | None = None

    def __call__(self, stream: BinaryIO) -> None:
        result = []
        for rawline in iter(stream.readline, b""):
            if self.capture:
                result.append(rawline[:-1])
            line = rawline.decode("utf-8")
            if line.strip().startswith("Warning:"):
                line = line.replace("Warning:", f"{COLORS['red']}Warning:{self.clear}{self.color}")
            print(
                f"{self.pre}{self.color}{line.rstrip()}{self.clear}",
                flush=True,
                file=sys.stderr,
            )
        stream.close()
        if self.capture:
            self.result = b"\n".join(result).decode(self.encode) if self.encode else b"\n".join(result)


def mkpaths(args: Paths) -> list[str]:
    return [str(args)] if isinstance(args, (str, Path)) else [str(a) for a in args]


def runc(
    args: Paths,
    stdout: OMode = "display",
    stderr: EMode = "display",
    overrides: dict[str, str] | None = None,
    timeout: datetime.datetime | datetime.timedelta | int | float | None = None,
    **kwargs,
) -> str | bytes | None:
    kwargs["env"] = kwargs.pop("env") if "env" in kwargs else os.environ.copy()
    kwargs["env"].update(overrides or {})
    kwargs["cwd"] = (str(v) if (v := kwargs.pop("cwd")) else None) if "cwd" in kwargs else None

    with subprocess.Popen(mkpaths(args), stderr=subprocess.PIPE, stdout=subprocess.PIPE, **kwargs) as process:
        if stdout == "capture":
            ofiltermap: BaseFilter = CaptureFilter()
        elif stdout == "null":
            ofiltermap = BaseFilter()
        elif stdout == "display":
            ofiltermap = DisplayFilter(COLORS["blue"], "   | ")
        elif stdout == "capture+display":
            ofiltermap = DisplayFilter(COLORS["blue"], "   | ", capture=True)
        elif isinstance(stdout, BaseFilter):
            ofiltermap = stdout
        else:
            raise RuntimeError(f"unsupported type in {stdout=}")
        othread = threading.Thread(target=ofiltermap, args=(process.stdout,), daemon=True)

        if stderr == "null":
            efiltermap: BaseFilter = BaseFilter()
        elif stderr == "display":
            efiltermap = DisplayFilter(COLORS["green"], "   | ")
        elif isinstance(stderr, BaseFilter):
            efiltermap = stderr
        else:
            raise RuntimeError(f"unsupported type in {stderr=}")
        ethread = threading.Thread(
            target=efiltermap,
            args=(process.stderr,),
            daemon=True,
        )
        expiry = None
        if timeout:
            if isinstance(timeout, datetime.datetime):
                expiry = timeout
            elif isinstance(timeout, datetime.timedelta):
                expiry = datetime.datetime.now() + timeout
            elif isinstance(timeout, (int, float)):
                expiry = datetime.datetime.now() + datetime.timedelta(seconds=float(timeout))

        started = False
        while True:
            if expiry and datetime.datetime.now() > expiry:
                process.terminate()
                break
            if not started:
                othread.start()
                ethread.start()
                started = True
            time.sleep(POLL_S)
            if process.poll() is not None:
                break

        if started:
            othread.join()
            ethread.join()

    kind = RunnerError
    if expiry and datetime.datetime.now() > expiry:
        kind = TimeoutRunnerError

    if process.returncode:
        # envs = " ".join(f'{k}="{v}"' for k, v in (overrides or {}).items())
        cmdline = subprocess.list2cmdline(mkpaths(args))
        raise kind(f"failed to execute {cmdline}", kwargs["cwd"], overrides, cmdline, process.returncode)
    return ofiltermap.result if hasattr(ofiltermap, "result") else None


if __name__ == "__main__":
    x = runc(["ls", "-l"], "capture", "display")
    print(x)



from __future__ import annotations

import random
import string
import sys
from io import StringIO
from pathlib import Path
from types import ModuleType
from typing import Any
from urllib.parse import urlparse
from urllib.request import urlopen


def loadmod(path: Path | str | StringIO, variables: dict[str, Any] | None = None, name: str = "") -> ModuleType:
    random_name = "".join(random.choices(string.ascii_lowercase, k=5))
    random_name = f"loadmod.{random_name}"

    txt = None
    if isinstance(path, Path):
        txt = path.read_text()
    elif isinstance(path, StringIO):
        txt = path.getvalue()
        path = "<string>"
    elif (parsed := urlparse(str(path))).scheme in {"http", "https"}:
        txt = str(urlopen(str(path)).read(), encoding="utf-8")
        path = f"<string {path}>"
    elif parsed.scheme in {"file"}:
        txt = Path(parsed.path).read_text()
        path = parsed.netloc
    else:
        raise ValueError(f"unsupported path type: {path}")

    mod = ModuleType(name)
    mod.__dict__.update(variables or {})
    mod.__dict__["__name__"] = name or random_name
    mod.__file__ = str(path)
    code_obj = compile(txt, filename=str(path), mode="exec")
    sys.modules[name] = mod
    exec(code_obj, mod.__dict__)
    return mod

    # spec = spec_from_file_location(Path(path).name, Path(path))
    # module = module_from_spec(spec)  # type: ignore
    # spec.loader.exec_module(module)  # type: ignore
    # return module


class NA:
    pass


def diffdict(
    left: dict[str, Any], right: dict[str, Any], exclude: list[str] | None = None, na: str | type[NA] = NA
) -> dict[str, tuple[Any, Any]]:
    result = {}
    for key in sorted(set(left) | set(right)):
        if exclude and key in exclude:
            continue
        if key not in left:
            result[key] = (na, right[key])
        elif key not in right:
            result[key] = (left[key], na)
        elif left[key] != right[key]:
            result[key] = (left[key], right[key])
    return result



import pytest


@pytest.fixture(scope="function")
def loader(resolver):
    from acbox.loader import LoaderBase

    yield LoaderBase([resolver.root.parent])


def test_base(resolver, loader):
    target = resolver.lookup("simple-script.py")
    found = loader.lookup("data/simple-script.py")
    assert target == found


def test_load_data(loader):
    assert {"A": 1, "B": {"C": 2, "D": 3}} == loader.load("data/sample.json")
    assert {"A": 1, "B": {"C": 2, "D": 3}} == loader.load("data/sample.yml")



import subprocess
import sys

import pytest


@pytest.fixture(scope="function")
def examples(resolver):
    yield lambda path: resolver.lookup(resolver.root.parent.parent / path)


@pytest.fixture(scope="function")
def runner(resolver):
    def run(example, args=None):
        script = resolver.lookup(resolver.root.parent.parent / example)
        p = subprocess.Popen(
            [sys.executable, str(script), *[str(c) for c in (args or [])]], encoding="utf-8", stderr=subprocess.PIPE, stdout=subprocess.PIPE
        )
        out, err = p.communicate()
        return p.returncode, out, err

    yield run


def test_single_command_help(runner):
    code, out, err = runner("docs/examples/single-command.py", ["--help"])
    assert (
        """\
usage: single-command.py [-h] [-v] [-q]

options:
  -h, --help     show this help message and exit

Logging:
  Logging related options

  -v, --verbose  report verbose logging (default: None)
  -q, --quiet    report quiet logging (default: None)
"""
        == out
    )


@pytest.mark.skip(reason="broken")
def test_single_command_run(runner):
    code, out, err = runner("docs/examples/single-command.py")
    assert not code
    assert (
        """\
args:
  .error: (callable) abort a script with an error message
  .modules: (<class 'list'>)
     <module 'acbox.cli.script' from '/Users/antonio/Projects/github/acbox/src/acbox/cli/script.py'>
     <module '__main__' from '/Users/antonio/Projects/github/acbox/docs/examples/single-command.py'>
"""
        == out
    )

    assert (
        """\
INFO:__main__:an info message, you can silence it with -q|--quiet
WARNING:__main__:a warning!
"""
        == err
    )



from __future__ import annotations

import dataclasses as dc
import logging
from pathlib import Path
from typing import Any, Literal, Sequence

logger = logging.getLogger(__name__)

Paths = str | Path | Sequence[str | Path]


def makepaths(args: Paths) -> list[Path]:
    if isinstance(args, (str, Path)):
        return [Path(args)]
    return [Path(a) for a in args]


@dc.dataclass
class LoaderBase:
    paths: list[Path] | None = None

    def __post_init__(self):
        self.paths = [Path(p).expanduser() for p in (self.paths or [])]

    def lookup(self, name: Path | str) -> Path | None:
        for item in self.paths or [Path.cwd()]:
            if (found := (item / name)).exists():
                return found
        return None

    def load(self, name: Path | str, mode: Literal["auto", "binary", "raw", "text"] = "auto") -> Any:
        if not (path := self.lookup(name)):
            raise FileNotFoundError(2, "cannot lookup name", name)

        if mode in {"binary", "raw"}:
            return path.read_bytes()
        elif mode == "text":
            return path.read_text()

        kind = path.suffix
        if kind.upper() in {".JSON"}:
            from json import loads

            return loads(self.load(path.absolute(), "text"))
        elif kind.upper() in {".YAML", ".YML"}:
            from yaml import safe_load

            return safe_load(self.load(path.absolute(), "text"))
        else:
            raise TypeError(f"cannot find type ({kind}) for {path}")
        return None



from __future__ import annotations

import argparse
import contextlib
import functools
import inspect
import logging
import sys
import time
from typing import Any, Callable

from .parsers.base import ArgumentParserBase
from .parsers.simple import ArgumentParser
from .shared import AbortCliError, AbortWrongArgumentError

log = logging.getLogger(__name__)


@contextlib.contextmanager
def setup(
    function: Callable,
    add_arguments: (Callable[[ArgumentParserBase], None] | Callable[[argparse.ArgumentParser], None] | None) = None,
    process_args: (Callable[[argparse.Namespace], argparse.Namespace | None] | None) = None,
):
    sig = inspect.signature(function)
    module = inspect.getmodule(function)

    # the cli decorated function might have two special parameters:
    #  - args this will receive non-parsed arguments (eg from nargs="*")
    #  - parser escape hatch to the parser object (probably never used)
    if "args" in sig.parameters and "parser" in sig.parameters:
        raise RuntimeError(f"function '{module}.{function.__name__}' cannot take args and parser at the same time")

    # doc is taken from the function itself or the containing module
    description, _, epilog = (function.__doc__ or module.__doc__ or "").strip().partition("\n")
    epilog = f"{description}\n{'-' * len(description)}\n{epilog}"
    description = ""

    # extract parser info/fallbacks from all these modules
    modules = [
        sys.modules[__name__],
    ]
    if module:
        modules.append(module)

    parser = ArgumentParser.get_parser(modules, description=description, epilog=epilog)
    if add_arguments and (callbacks := add_arguments(parser)):
        if isinstance(callbacks, list):
            parser.callbacks.extend(callbacks)
        else:
            parser.callbacks.append(callbacks)

    kwargs = {}
    if "parser" in sig.parameters:
        kwargs["parser"] = parser

    t0 = time.monotonic()
    success = "completed"
    errormsg = ""
    show_timing = True
    try:
        if "parser" not in sig.parameters:
            args = parser.parse_args()
            if process_args:
                args = process_args(args) or args

            if "args" in sig.parameters:
                kwargs["args"] = args
        yield sig.bind(**kwargs)

    except argparse.ArgumentError:
        sys.exit(2)
        pass
    except AbortCliError as exc:
        show_timing = False
        if exc.args:
            print(str(exc), file=sys.stderr)
        sys.exit(2)
    except AbortWrongArgumentError as exc:
        show_timing = False
        parser.print_usage(sys.stderr)
        print(f"{parser.prog}: error: {exc.args[0]}", file=sys.stderr)
        sys.exit(2)
    except SystemExit as exc:
        show_timing = False
        sys.exit(exc.code)
    except Exception:
        log.exception("un-handled exception")
        success = "failed with an exception"
    finally:
        if show_timing:
            delta = round(time.monotonic() - t0, 2)
            log.debug("task %s in %.2fs", success, delta)
    if errormsg:
        parser.error(errormsg)


def cli(
    add_arguments: (Callable[[ArgumentParserBase], Any] | Callable[[argparse.ArgumentParser], Any] | None) = None,
    process_args: (Callable[[argparse.Namespace], argparse.Namespace | None] | None) = None,
):
    def _cli1(function):
        module = inspect.getmodule(function)

        if inspect.iscoroutinefunction(function):

            @functools.wraps(function)
            async def _cli2(*args, **kwargs):
                with setup(function, add_arguments, process_args) as ba:
                    return await function(*ba.args, **ba.kwargs)

        else:

            @functools.wraps(function)
            def _cli2(*args, **kwargs):
                with setup(function, add_arguments, process_args) as ba:
                    return function(*ba.args, **ba.kwargs)

        _cli2.attributes = {
            "doc": function.__doc__ or module.__doc__ or "",
        }
        return _cli2

    return _cli1



def test_script_lookup(resolver):
    path = resolver.lookup("simple-script.py")
    assert path
    assert path.exists()


def test_script_load(resolver):
    data = resolver.load("simple-script.py", "text")
    assert (
        data[:525]
        == """
# call this:
#   simple-script.py - writes simple messages to stdout/stderr
#   VALUE=123 simple-script.py - same as above, but writes VALUE message too
#   WAIT=2 VALUE=123 simple-script.py - same as above, but wait 2s before printing last message
#   ABORT=12 simple-script.py - same as above, but abort the script with error code 12
import os
import sys
import time
from typing import TextIO


def hello(index: int, msg: str, file: TextIO = sys.stdout) -> None:
    print(f"{index}, got '{msg}' ({file.name})", file=file)
""".lstrip()
    )

    mod = resolver.load("simple-script.py", "mod")
    assert callable(mod.hello)



# 1. info tool
# 2. replacer
import logging
import sys
from pathlib import Path

import click

from acbox.clickwrapper import Namespace, clickwrap, group

log = logging.getLogger(__name__)


@group()
def main():
    print("HELLO")


@main.command(aliases=["i", "info"])
@clickwrap("default")
def info(args: Namespace) -> None:
    "prints info data"
    from acbox.toolbox.info import main

    sys.exit(main())


def validate_define(ctx, param, values):
    if not values:
        return values
    for value in values:
        if value.count("=") != 1:
            raise click.BadParameter(f"{param.opts} must be in KEY=VALUE format: {value}")
    return dict(tuple(value.split("=")) for value in values)


@main.command(aliases=["r"])
@clickwrap("default")
@click.option("-D", "--define", callback=validate_define, multiple=True)
@click.option("-i", "--inplace", is_flag=True)
@click.option("-n", "--dry-run", is_flag=True)
@click.argument("files", nargs=-1)
def replace(args: Namespace) -> None:
    from acbox.toolbox.replacer import fix, process

    if not args.files:
        return
    if len(args.files) == 1 and not args.inplace:
        print(process(Path(args.files[0]).read_text(), defines=args.define))
        return

    if not args.inplace:
        args.error("with multiple sources, you need the -i|--inplace flag")

    for path in args.files:
        log.info("processing %s", path)
        fix(Path(path), defines=args.define, inplace=args.inplace)
